# SIS as a compartmented model
#
# Copyright (C) 2017--2020 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import CompartmentedModel
import sys
if sys.version_info >= (3, 8):
    from typing import Final, Dict, Any
else:
    # backport compatibility with older typing
    from typing import Dict, Any
    from typing_extensions import Final


class SIS(CompartmentedModel):
    '''The Susceptible-Infected-Susceptible :term:`compartmented model of disease`.
    Susceptible nodes are infected by infected neighbours, and recover back
    to the susceptible state (which allows future re-infection, unlike for
    :class:`SIR`.'''

    # Model  parameters
    P_INFECTED : Final[str] = 'epydemic.SIS.pInfected'  #: Parameter for probability of initially being infected.
    P_INFECT : Final[str] = 'epydemic.SIS.pInfect'      #: Parameter for probability of infection on contact.
    P_RECOVER : Final[str] = 'epydemic.SIS.pRecover'    #: Parameter for probability of recovery (returning to susceptible).

    # Possible dynamics states of a node for SIR dynamics
    SUSCEPTIBLE : Final[str] = 'epydemic.SIS.S'         #: Compartment for nodes susceptible to infection.
    INFECTED : Final[str] = 'epydemic.SIS.I'            #: Compartment for nodes infected.

    # Locus containing the edges at which dynamics can occur
    SI : Final[str] = 'epydemic.SIS.SI'                 #: Edge able to transmit infection.

    def __init__(self):
        super(SIS, self).__init__()

    def build(self, params :Dict[str, Any]):
        '''Build the SIS model.

        :param params: the model parameters'''
        super(SIS, self).build(params)
        
        pInfected = params[self.P_INFECTED]
        pInfect = params[self.P_INFECT]
        pRecover = params[self.P_RECOVER]

        self.addCompartment(self.SUSCEPTIBLE, 1 - pInfected)
        self.addCompartment(self.INFECTED, pInfected)

        self.trackEdgesBetweenCompartments(self.SUSCEPTIBLE, self.INFECTED, name = self.SI)
        self.trackNodesInCompartment(self.INFECTED)

        self.addEventPerElement(self.INFECTED, pRecover, self.recover)
        self.addEventPerElement(self.SI, pInfect, self.infect)

    def infect(self, t : float, e : Any):
        '''Perform an infection event. This changes the compartment of
        the susceptible-end node to :attr:`INFECTED`. It also marks the edge
        traversed as occupied.

        :param t: the simulation time
        :param e: the edge transmitting the infection, susceptible-infected'''
        (n, _) = e
        self.changeCompartment(n, self.INFECTED)
        self.markOccupied(e, t)

    def recover(self, t : float, n : Any):
        '''Perform a recovery event. This changes the compartment of
        the node back to :attr:`SUSCEPTIBLE`, allowing re-infection.

        :param t: the simulation time (unused)
        :param n: the node'''
        self.changeCompartment(n, self.SUSCEPTIBLE)
    
                
   



# Bond percolation process
#
# Copyright (C) 2020 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import Process, Edge
import numpy                      # type: ignore
import sys
if sys.version_info >= (3, 8):
    from typing import Final, Dict, Any, List
else:
    # backport compatibility with older typing
    from typing import Dict, Any, List
    from typing_extensions import Final


class Percolate(Process):
    '''A process that bond percolates a network with a given occupation probability. This
    can be used independently as a basis for percolation experiments, or as a precursor
    to other processes that need to run on a percolated network.
    
    Note that the percolation process defines no events: it performs percolation during
    the :meth:`build` method.'''
    
    # Experimental parameters
    T : Final[str] = 'epydemic.percolate.T'   #: Experimental parameter for the occupation probability of an edge.
    
    
    def __init__(self):
        super(Percolate, self).__init__()
    
    def percolate(self, T : float) -> List[Edge]:
        '''Percolate the network. Edges are retained ("occupied") with probability :math:`T`,
        with other edges being "unoccupied" with probability :math:`(1 - T)`.
        
        :param params: the experimental parameters
        :returns: the set of edges to be removed, i.e., those that are unoccupied'''
        rng = numpy.random.default_rng()
        g = self.network()
        
        # percolate the network
        unoccupied = []
        for e in g.edges():
            # edges are occupied with probability T
            if rng.random() > T:
                # retain unoccupied edges
                unoccupied.append(e)
                
        # return the unoccupied edges
        return unoccupied
        
    def unoccupy(self, unoccupied : List[Edge]):
        '''Handle edges, given the set that is left unoccupied by the percolation
        process. The default removes the unoccupied edges. This can be overridden to
        change the action.
        
        :param unoccupied: set of unoccupied'''''
        g = self.network()
        g.remove_edges_from(unoccupied)

    def build(self, params : Dict[str, Any]):
        '''Percolate the network.
        
        :param params: the experimental parameters'''
        super(Percolate, self).build(params)
        
        # percolate the network
        T = params[self.T]
        unoccupied = self.percolate(T)
        self.unoccupy(unoccupied)




# SIR as a compartmented model
#
# Copyright (C) 2017--2020 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import CompartmentedModel
import sys
if sys.version_info >= (3, 8):
    from typing import Final, Dict, Any
else:
    # backport compatibility with older typing
    from typing import Dict, Any
    from typing_extensions import Final

class SIR(CompartmentedModel):
    '''The Susceptible-Infected-Removed :term:`compartmented model of disease`.
    Susceptible nodes are infected by infected neighbours, and recover to
    removed.'''
    
    # Model parameters
    P_INFECTED : Final[str] = 'epydemic.SIR.pInfected'  #: Parameter for probability of initially being infected.
    P_INFECT : Final[str] = 'epydemic.SIR.pInfect'      #: Parameter for probability of infection on contact.
    P_REMOVE : Final[str] = 'epydemic.SIR.pRemove'      #: Parameter for probability of removal (recovery).
    
    # Possible dynamics states of a node for SIR dynamics
    SUSCEPTIBLE : Final[str] = 'epydemic.SIR.S'         #: Compartment for nodes susceptible to infection.
    INFECTED : Final[str] = 'epydemic.SIR.I'            #: Compartment for nodes infected.
    REMOVED : Final[str] = 'epydemic.SIR.R'             #: Compartment for nodes recovered/removed.

    # Locus containing the edges at which dynamics can occur
    SI : Final[str] = 'epydemic.SIR.SI'                 #: Edge able to transmit infection.

    def __init__(self):
        super(SIR, self).__init__()

    def build(self, params : Dict[str, Any]):
        '''Build the SIR model.

        :param params: the model parameters'''
        super(SIR, self).build(params)

        pInfected = params[self.P_INFECTED]
        pInfect = params[self.P_INFECT]
        pRemove = params[self.P_REMOVE]

        self.addCompartment(self.SUSCEPTIBLE, 1 - pInfected)
        self.addCompartment(self.INFECTED, pInfected)
        self.addCompartment(self.REMOVED, 0.0)

        self.trackEdgesBetweenCompartments(self.SUSCEPTIBLE, self.INFECTED, name=self.SI)
        self.trackNodesInCompartment(self.INFECTED)

        self.addEventPerElement(self.SI, pInfect, self.infect)
        self.addEventPerElement(self.INFECTED, pRemove, self.remove)

    def infect(self, t : float, e : Any):
        '''Perform an infection event. This changes the compartment of
        the susceptible-end node to :attr:`INFECTED`. It also marks the edge
        traversed as occupied.

        :param t: the simulation time
        :param e: the edge transmitting the infection, susceptible-infected'''
        (n, _) = e
        self.changeCompartment(n, self.INFECTED)
        self.markOccupied(e, t)

    def remove(self, t : float, n : Any):
        '''Perform a removal event. This changes the compartment of
        the node to :attr:`REMOVED`.

        :param t: the simulation time (unused)
        :param n: the node'''
        self.changeCompartment(n, self.REMOVED)
    
                
   



# Test basic process functions
#
# Copyright (C) 2020 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import Process, Dynamics
import unittest
import networkx

class ProcessTest(unittest.TestCase):

    def setUp(self):
        self._g = networkx.Graph()
        self._p = Process()
        self._dyn = Dynamics(self._p)
        self._p.reset()
        self._p.setNetwork(self._g)
        self._p.build(dict())
        self._p.setUp(dict())

    def testAddNode(self):
        '''Test all node addition methods pass-though correctly.'''

        # test single addition
        self._p.addNode(1)
        self.assertCountEqual(self._g.nodes(), [ 1 ])

        # test bulk addition
        self._p.addNodesFrom([ 2, 3, 4 ])
        self.assertCountEqual(self._g.nodes(), [ 1, 2, 3, 4 ])

        # test repeated addiion does nothing
        self._p.addNode(1)
        self.assertCountEqual(self._g.nodes(), [ 1, 2, 3, 4 ])

        # test attributes are stored correctly
        self._p.addNode(5, test=1)
        self.assertCountEqual(self._g.nodes(), [ 1, 2, 3, 4, 5 ])
        self.assertEqual(self._g.nodes[5]['test'], 1)
        self._p.addNodesFrom([ 6, 7 ], test=5)
        self.assertEqual(self._g.nodes[6]['test'], 5)
        self.assertEqual(self._g.nodes[7]['test'], 5)

    def testAddEdge(self):
        '''Test all the ewdge addition methods pass-through correctly.'''
        self._p.addNodesFrom([ 1, 2, 3, 4, 5 ])

        # test single addition
        self._p.addEdge(1, 2)
        self.assertCountEqual(self._g.edges(), [ (1, 2) ])

        # test bulk addition
        self._p.addEdgesFrom([ (1, 3), (3, 4), (4, 5) ])
        self.assertCountEqual(self._g.edges(), [ (1, 2), (1, 3), (3, 4), (4, 5) ])

        # test repeated addition does nothing
        self._p.addEdge(1, 2)
        self.assertCountEqual(self._g.edges(), [ (1, 2), (1, 3), (3, 4), (4, 5) ])

        # test attributes are stored correctly
        self._p.addEdge(1, 5, test=1)
        self.assertCountEqual(self._g.edges(), [ (1, 5), (1, 2), (1, 3), (3, 4), (4, 5) ])
        self.assertEqual(self._g.edges[1, 5]['test'], 1)
        self._p.addEdgesFrom([ (2, 5), (3, 5) ], test=5)
        self.assertEqual(self._g.edges[2, 5]['test'], 5)
        self.assertEqual(self._g.edges[3, 5]['test'], 5)

        # test adding an edge with a missing endpoint fails
        with self.assertRaises(Exception):
            self._p.addEdge(1, 7)
        with self.assertRaises(Exception):
            self._p.addEdge(7, 1)


if __name__ == '__main__':
    unittest.main()



# Test network dynamics
#
# Copyright (C) 2017--2019 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *
import unittest

class NetworkDynamicsTest(unittest.TestCase):

    def pendingEvents(self, dyn, t):
        '''Return all the events posted to occur at or before time t.

        :param dyn: the dynamics
        :param t: the time
        :returns: the events'''
        evs = []
        while True:
            ev = dyn.nextPendingEventBefore(t)
            if ev is None:
                break
            else:
                evs.append(ev)
        return evs

    def testPosting( self ):
        '''Test posting events.'''
        p = Process()
        dyn = Dynamics(p)

        # event function builder, will increment the shared
        # value when fired
        self._v = 0
        def make_ef( w ):
            def ef( t, e ):
                self._v = self._v + w
            return ef
        
        # post some events at different times
        dyn.postEvent(1, None, make_ef(1))
        dyn.postEvent(2, None, make_ef(20))
        dyn.postEvent(3, None, make_ef(100))
        dyn.postEvent(4, None, make_ef(200))

        # check no events before the first one posted
        self.assertCountEqual(self.pendingEvents(dyn, 0.5), [])

        # check firing of earliest event
        pefs = self.pendingEvents(dyn, 1)
        self.assertTrue(len(pefs), 1)
        (pefs[0])()                # fire the event
        self.assertTrue(self._v, 1)
        self.assertCountEqual(self.pendingEvents(dyn, 1), [])

        # check multiple events coming off in the right order
        pefs = self.pendingEvents(dyn, 3)
        self.assertTrue(len(pefs), 2)
        (pefs[0])()
        self.assertTrue(self._v, 21)
        (pefs[1])()
        self.assertTrue(self._v, 121)
        self.assertCountEqual(self.pendingEvents(dyn, 3), [])

        # check we can run all remaining events
        dyn.runPendingEvents(10)       # a long time in the future
        self.assertTrue(self._v, 321)
        self.assertCountEqual(self.pendingEvents(dyn, 10), [])
        self.assertCountEqual(self.pendingEvents(dyn, 20), [])
        
    def testPostedPosting( self ):
        '''Test the case when a posted event itself posts an event.'''
        p = Process()
        dyn = Dynamics(p)

        # event function builder, will increment the shared
        # value when fired and also post a further event
        self._v = 0
        def make_ef(dt):
            def ef(t, e):
                self._v = self._v + 1
                p.postEvent(t + dt, None, make_ef(dt))
            return ef
        
        # post the first of the events
        p.postEvent(1, None, make_ef(5))

        # post an intermediate event that should be fired alongside a regular one
        def inter(t, e):
            self._v = self._v + 1000
        p.postEvent(12, None, inter)

        # carefully run the event sequence for a while
        for t in range(1, 25, 5):
            nev = dyn.runPendingEvents(t)
            v = self._v
            if t == 1:
                self.assertEqual(v, 1)
                self.assertEqual(nev, 1)
            if t == 6:
                self.assertEqual(v, 2)
                self.assertEqual(nev, 1)
            if t == 11:
                self.assertEqual(v, 3)
                self.assertEqual(nev, 1)
            if t == 16:
                self.assertEqual(v, 1004)
                self.assertEqual(nev, 2)
            if t == 21:
                self.assertEqual(v, 1005)
                self.assertEqual(nev, 1)

        # should be one event left
        self.assertEqual(len(self.pendingEvents(dyn, 100)), 1)
        
    def testRepeating(self):
        '''Test that repeating events repeat properly.'''
        p = Process()
        dyn = Dynamics(p)

        # event function builder to count the times
        self._ps = []
        def make_ef():
            def ef(t, e):
                self._ps.append((t))
            return ef

        # post a repeating event
        p.postRepeatingEvent(1, 3, None, make_ef())
        for t in range(20):
            dyn.runPendingEvents(t)

        # should be one event left
        self.assertEqual(len(self.pendingEvents(dyn, 100)), 1)

        # check list of times
        self.assertEqual(self._ps, [1, 4, 7, 10, 13, 16, 19])

    def testNoLoci(self):
        '''Test we can handle processes without loci.'''
        p = Process()
        dyn = Dynamics(p)
        self.assertCountEqual(dyn.lociForProcess(p), dict())

if __name__ == '__main__':
    unittest.main()



# Test process monitoring
#
# Copyright (C) 2020 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *
import unittest
import epyc
import networkx

class MonitoredSIR(SIR, Monitor):
    
    def __init__(self):
        super(MonitoredSIR, self).__init__()


class MonitorTest(unittest.TestCase):

    def testSimple( self ):
        '''Test we capture the right time series.'''
        m = MonitoredSIR()
        m.setMaximumTime(100)
        e = StochasticDynamics(m, networkx.erdos_renyi_graph(1000, 5.0 / 1000))

        param = dict()
        param[SIR.P_INFECTED] = 0.01
        param[SIR.P_INFECT] = 0.002
        param[SIR.P_REMOVE] = 0.002
        param[Monitor.DELTA] = 1.0

        rc = e.set(param).run()
        self.assertIn(Monitor.TIMESERIES, rc[epyc.Experiment.RESULTS])
        self.assertSetEqual(set(rc[epyc.Experiment.RESULTS].keys()), set([Monitor.TIMESERIES, SIR.SUSCEPTIBLE, SIR.INFECTED, SIR.REMOVED]))
        self.assertSetEqual(set(rc[epyc.Experiment.RESULTS][Monitor.TIMESERIES].keys()), set([Monitor.OBSERVATIONS, SIR.SI, SIR.INFECTED]))
        # the next test is >=, not =, because some events may be drawn after the maxiumum time,
        # but the time is short enough that the number of infecteds won't be exhausted beforehand
        self.assertGreaterEqual(len(rc[epyc.Experiment.RESULTS][Monitor.TIMESERIES][Monitor.OBSERVATIONS]), 100)
        n = len(rc[epyc.Experiment.RESULTS][Monitor.TIMESERIES][Monitor.OBSERVATIONS])
        for k in [SIR.SI, SIR.INFECTED]:
            self.assertEqual(len(rc[epyc.Experiment.RESULTS][Monitor.TIMESERIES][k]), n)

if __name__ == '__main__':
    unittest.main()


# Test the Newman-Ziff percolation algorithms
#
# Copyright (C) 2017--2021 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import BondPercolation, SitePercolation, PLCNetwork
from epyc import Lab, RepeatedExperiment
import numpy
import unittest
from networkx import complete_graph

class NewmanZiffTest(unittest.TestCase):

    # ---------- Bond percolation ----------

    def testBondZero(self):
        '''Test we can do the basic bond percolation operations.'''
        g = complete_graph(5)
        e = BondPercolation(g)
        _ = e.run()
    
    def testBondTransition(self):
        '''Test the bond percolation transition on a PLC network.'''
        params = dict()
        params[PLCNetwork.N] = 1000
        params[PLCNetwork.EXPONENT] = 2.5
        params[PLCNetwork.CUTOFF] = 20
        e = BondPercolation(PLCNetwork())
        rc = e.set(params).run()
        print(rc)

    def testResultLayout(self):
        '''Test the results land correctly in the notebook.'''
        lab = Lab()
        lab[PLCNetwork.N] = 1000
        lab[PLCNetwork.EXPONENT] = 2.5
        lab[PLCNetwork.CUTOFF] = 20
        e = BondPercolation(PLCNetwork(), samples=50)
        lab.runExperiment(e)
        df = lab.notebook().current().dataframe()
        self.assertEqual(len(df), 50)
        self.assertIn(0.0, df[BondPercolation.P])
        self.assertIn(1.0, df[BondPercolation.P])

    def testSamplePoints(self):
        '''Test we can set explicit sample points.'''
        lab = Lab()
        lab[PLCNetwork.N] = 1000
        lab[PLCNetwork.EXPONENT] = 2.5
        lab[PLCNetwork.CUTOFF] = 20
        samples = [0.0, 0.5, 1.0]
        e = BondPercolation(PLCNetwork(), samples=samples)
        lab.runExperiment(e)
        df = lab.notebook().current().dataframe()
        self.assertEqual(len(df), 3)
        points = sorted(list(df[BondPercolation.P]))
        self.assertEqual(points[0], 0.0)
        self.assertAlmostEqual(points[1], 0.5, places=2)
        self.assertEqual(points[2], 1.0)

    def testRepeats(self):
        '''Test we play well with repeated experiments.'''
        lab = Lab()
        lab[PLCNetwork.N] = 1000
        lab[PLCNetwork.EXPONENT] = 2.5
        lab[PLCNetwork.CUTOFF] = 20
        e = BondPercolation(PLCNetwork(), samples=10)
        lab.runExperiment(RepeatedExperiment(e, 3))
        df = lab.notebook().current().dataframe()
        self.assertEqual(len(df), 3 * 10)
        for p in df[BondPercolation.P]:
            self.assertEqual(len(df[df[BondPercolation.P] == p]), 3)


   # ---------- Site percolation ----------
    
    def testSiteZero(self):
        '''Test we can do the basic site percolation operations.'''
        g = complete_graph(5)
        e = SitePercolation(g)
        _ = e.run()

    def testSiteTransition(self):
        '''Test the site percolation transition on a PLC network.'''
        params = dict()
        params[PLCNetwork.N] = 1000
        params[PLCNetwork.EXPONENT] = 2.5
        params[PLCNetwork.CUTOFF] = 20
        e = SitePercolation(PLCNetwork())
        rc = e.set(params).run()
        print(rc)    


if __name__ == '__main__':
    unittest.main()



# Monitor the progress of an epidemic
#
# Copyright (C) 2020 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import Process
import sys
if sys.version_info >= (3, 8):
    from typing import Final, Dict, Any
else:
    # backport compatibility with older typing
    from typing import Dict, Any
    from typing_extensions import Final

class Monitor(Process):
    '''Add progress monitoring to a process. This class captures the
    sizes of all loci in the process at regular intervals, returning
    them as time series.'''
    
    # Experimental parameters
    DELTA : Final[str] = "epydemic.Monitor.time_delta"           #: Parameter for the time interval for observations.
    
    # Results
    OBSERVATIONS : Final[str] = 'epydemic.Monitor.observations'  #: Result holding the times of the observations.
    TIMESERIES : Final[str] = "epydemic.Monitor.timeseries"      #: Result holding a dict mapping locus names to a dict of sample time and size.
 

    def __init__(self):
        super(Monitor, self).__init__()

    def reset(self):
        '''Reset the process.'''
        super(Monitor, self).reset()
        self._timeSeries = None
        
    def build(self, params : Dict[str, Any]):
        '''Build the observation process.
        
        :param params: the experimental parameters'''
        super(Monitor, self).build(params)
 
        # post a repeating event to observe the process
        delta = params[self.DELTA]
        self.postRepeatingEvent(0.0, delta, None, self.observe)
        
    def observe(self, t : float, e : Any):
        '''Observe the network, capturing the sizes of all loci which are then
        stored into individual time series. Another time series, tagged 
        :attr:`OBSERVATIONS`, is created to store the sequence of times at which observations
        were made.
        
        :param t: the current simulation time
        :param e: the element (ignored)'''
 
        # if this is the first run, build the initial dicts: one
        # per locus and one for the observation times
        # (We can't do this as part of build() as it depends on which
        # other processes we're composed with, and in what order)
        if self._timeSeries is None:
            self._timeSeries = dict()
            self._timeSeries[self.OBSERVATIONS] = []
            for n in self.dynamics().loci().keys():   # loci for all processes in this experiment
                self._timeSeries[n] = []

        # make the observation
        self._timeSeries[self.OBSERVATIONS].append(t)
        for (n, l) in self.loci().items():
            self._timeSeries[n].append(len(l))
        
    def results(self) -> Dict[str, Any]:
        '''Return the time series as a dict tagged :attr:`TIMESERIES`. There is
        one time series *per* locus, plus one for the sequence of times at which the
        observations were made.
        
        :returns: the results'''
        rc = super(Monitor, self).results()
        
        # store the series
        rc[self.TIMESERIES] = self._timeSeries
        
        return rc


# Test percolation process
#
# Copyright (C) 2017--2020 Simon Dobson
#
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *
import networkx
import numpy
import unittest

class CaptureResidualNetwork(StochasticDynamics):
    '''Class that captures the residual network after running,
    to make tests simpler.'''

    def __init__(self, p, g):
        super(CaptureResidualNetwork, self).__init__(p, g)

    def tearDown(self):
        ''' Store the resulting network after running.'''
        self._residual = self.network()
        super(CaptureResidualNetwork, self).tearDown()


class PercolationTest(unittest.TestCase):

    def testProduceER(self):
        '''Test that a percolation process on a complete network creates an ER network.'''
        N = 2000
        kmean = 15
        phi =  kmean / N

        # create and percolate the complete graph
        g = networkx.complete_graph(N)
        params = dict()
        params[Percolate.T] = phi
        p = Percolate()
        e = CaptureResidualNetwork(p, g)
        rc = e.set(params).run()

        # check the topology of the resulting network
        rg = e._residual
        self.assertEqual(rg.order(), g.order())
        rg_kmean = numpy.mean(list(map(lambda nd: nd[1], list(rg.degree()))))
        self.assertAlmostEqual(rg_kmean, kmean, places=0)


if __name__ == '__main__':
    unittest.main()



# Networks experiment base class
#
# Copyright (C) 2017--2021 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epyc import Experiment
from epydemic import NetworkGenerator, FixedNetwork
from networkx import Graph
from typing import Union, Dict, Any, cast


class NetworkExperiment(Experiment):
    '''A very lightweight base class for providing a network
    to an `epyc` experiment. The network can either be a fixed network
    used for each experimental run, or a network generator that will be 
    used to generate a new instance for each run.
    
    :param g: (optional) prototype network or network generator'''
    def __init__(self, g : Union[Graph, NetworkGenerator] =None):
        super(NetworkExperiment, self).__init__()

        # turn a literal network into a network generator
        if isinstance(g, Graph):
            g = FixedNetwork(g)
        self._generator = cast(NetworkGenerator, g)       # network generator
        self._graph : Graph = None                        # working network instance


    # ---------- Configuration ----------

    def network(self) -> Graph:
        '''Return the network this dynamics is running over. This will return None
        unless we're actually running a simulation.

        :returns: the network'''
        return self._graph

    def setNetworkGenerator(self, g : Union[Graph, NetworkGenerator]):
        '''Set the network or generator for the networks the dynamics will run over.
        If a network is supplied rather than a generator it will be treated as an
        instance of :class:`FixedNetwork`.

        :param g: network or network generator'''
        if isinstance(g, Graph):
            g = FixedNetwork(g)
        self._generator = g

    def networkGenerator(self) -> NetworkGenerator:
        '''Return the generator used for this dynamics.

        :returns: the generator'''
        return self._generator


    # ---------- Set-up and tear-down ----------

    def setUp(self, params : Dict[str, Any]):
        '''Set up the experiment for a run. This creates a working copy
        of the network (class) underlying the experiment.

        :params params: the experimental parameters'''
        super(NetworkExperiment, self).setUp(params)

        # generate a working network instance
        self._graph = self.networkGenerator().set(params).generate()

    def tearDown(self):
        '''At the end of each experiment, throw away the working network.'''
        super(NetworkExperiment, self).tearDown()
        self._graph = None




# Networks process base class
#
# Copyright (C) 2017--2020 Simon Dobson
#
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import Locus, Node, Edge, Element
from networkx import Graph
from typing import Dict, List, Tuple, Any, Callable, Iterable, Union

# There is a circular import between Process and Dynamics at the typing level
# (but not at the execution level), when providing types for dynamics() and
# setDynamics(). To deal with this we only import Dynamics in order to
# type-check Process, and not for execution.
# (See https://www.stefaanlippens.net/circular-imports-type-hints-python.html)
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from epydemic import Dynamics

# Types for handling events
EventFunction = Callable[[float, Element], None]              #: Type of event-handler functions.
EventDistribution = List[Tuple[Locus, float, EventFunction]]  #: Type of event distributions.

class Process(object):
    """A process that runs over a network. This is the abstract base class for all network processes.
    It provides the essential routines to build, set-up, run, and extract results from a
    process. Sub-classes provide the actual behaviour by defining simulation events and
    attaching them to the network in different ways.

    A process definition is largely declarative, in the sense that it sets up event handlers
    and their probabilities. These are used by a process dynamics (a sub-class of :class:`Dynamics`)
    to actually run a simulation of the process.

    The process class includes an interface for interacting with the working network. The basic
    interface is extended and overridden in different sub-classes that interact with the network
    in different ways.
    """

    # defaults
    DEFAULT_MAX_TIME = 20000      #: Default maximum simulation time.

    def __init__(self):
        super(Process, self).__init__()

        # set the default maximum time, which persists across runs of the process
        self._maxTime = self.DEFAULT_MAX_TIME

        # reset this instance
        self.reset()


    # ---------- Setup and initialisation ----------

    def reset(self):
        """Reset the process ready to be built. This resets all the internal process state
        variables. Sub-classes should call the base method to make sure that the event
        sub-system is properly reset."""
        self._g = None

    def build(self, params : Dict[str, Any]):
        """Build the process model. This should be overridden by sub-classes, and should
        create the various elements of the model.

        :param params: the model parameters"""
        pass

    def setUp(self, params : Dict[str, Any]):
        """Set up the network under the given dynamics. The default does
        nothing: sub-classes should override it to initialise node states
        or establish other properties ready for the experiment.

        :param params: the simulation parameters"""
        pass

    def tearDown(self):
        """Tear down any structures built for this run of the process. The default does
        nothing."""
        pass

    def setNetwork(self, g : Graph):
        """Set the network the process is running over.

        :param g: the network"""
        self._g = g

    def network(self) -> Graph:
        """Return the network the process is running over.

        :returns: the network"""
        return self._g

    def setDynamics(self, d : 'Dynamics'):
        '''Set the instance of :class:`Dynamics` that runs the process.

        :param d: the dynamics'''
        self._dynamics = d

    def dynamics(self) -> 'Dynamics':
        '''Return the instance of :class:`Dynamics` running this process.

        :returns: the dynamics'''
        return self._dynamics

    def setMaximumTime(self, t : float):
        """Set the maximum default simulation time. The default is given by :attr:`DEFAULT_MAX_TIME`.
        This is used by :meth:`atEquilibrium` as the default way to determine equilibrium.

        The maximum time may be slightly exceeded due to the ways in which events are drawn. 

        Setting the maximum time persists across runs of the process, and isn't reset to its
        default by calling :meth:`reset`.

        :param t: the maximum simulation time"""
        self._maxTime = t

    def maximumTime(self) -> float:
        """Return the maximum assumed simulation time.

        :returns: the maximum simulation time"""
        return self._maxTime


   # ---------- Termination and results ----------

    def atEquilibrium(self, t : float) -> bool:
        """Test whether the process is an equilibrium. The default simply checks whether
        the simulation time exceeds the maximum given by :meth:`setMaximumTime`, which
        defaults to an arbitrary value given by :attr:`DEFAULT_MAX_TIME`.
        Sub-classes may override this to provide a more sensible definition: in many
        cases it makes sense to keep calling this method as part of any more advanced
        test, so as to cut-off runaway processes.

        :param t: the current simulation time
        :returns: True if the proceess is now at equilibrium"""
        return (t >= self.maximumTime())

    def results(self) -> Dict[str, Any]:
       """Create  and return an empty dict to be filled with experimental results.
       Sub-classes should extend this method to add results to the dict.

       :returns: an empty dict for experimental results"""
       return dict()


    # ---------- Accessing and evolving the network ----------

    def addNode(self, n : Node, **kwds):
        """Add a node to the working network. Any keyword arguments added as node attributes

        :param n: the new node
        :param kwds: (optional) node attributes"""
        self.network().add_node(n, **kwds)

    def addNodesFrom(self, ns : Iterable[Node], **kwds):
        """Add all the nodes in the given iterable to the working network. Any keyword arguments are
        added as node attributes. This works by calling :meth:`addNode` for each
        element of the iterable.

        :param ns: an iterable collection of nodes
        :param kwds: (optional) node attributes"""
        for n in ns:
            self.addNode(n, **kwds)

    def removeNode(self, n : Node):
        """Remove a node from the working network.

        :param n: the node"""
        self.network().remove_node(n)

    def removeNodesFrom(self, ns : Iterable[Node]):
        """Remove all the nodes in the given iterable froim the working network. This works by
        iteratively calling :meth:`removeNode` for each element of the iterable collection.

        :param ns: an iterable collection of nodes"""
        for n in ns:
            self.removeNode(n)

    def addEdge(self, n : Node, m : Node, **kwds):
        """Add an edge between nodes. Any keyword arguments are added as edge attributes.
        If the endpoint nodes do not already exist then an exception is raised.

        :param n: the start node
        :param m: the end node
        :param kwds: (optional) edge attributes"""
        g = self.network()
        if n not in g:
            raise Exception('No node {n} in network'.format(n = n))
        if m not in g:
            raise Exception('No node {n} in network'.format(n = m))
        g.add_edge(n, m, **kwds)

    def addEdgesFrom(self, es : Iterable[Edge], **kwds):
        """Add all the edges in the given iterable to the working network. Any keyword arguments are
        added as node attributes. This works by calling :meth:`addEdge` for each
        element of the iterable.

        :param es: an iterable collection of edges
        :param kwds: (optional) node attributes"""
        for e in es:
            (n, m) = e
            self.addEdge(n, m, **kwds)

    def removeEdge(self, n : Node, m : Node):
        """Remove an edge from the working network.

        :param n: the start node
        :param m: the end node"""
        self.network().remove_edge(n, m)

    def removeEdgesFrom(self, es : Iterable[Edge]):
        """Remove all the edges in the given iterable collection from the working network. This works
        by iteratively calling :meth:`removeEdge` for each edge in the iterable.

        :param es: an iterable collection of edges"""
        for e in es:
            self.removeEdge(*e)


    # ---------- Probabilistic and posted events ----------
    # These are helper methods that delegate to the dynamics

    def addLocus(self, n : str, l : Locus =None) -> Locus:
        """Add a named locus.

        :param n: the locus name
        :param l: the locus (defaults to a simple set-based locus)
        :returns: the locus"""
        return self._dynamics.addLocus(self, n, l)

    def loci(self) -> Dict[str, Locus]:
        '''Return the names of the loci that this process added.

        :returns: a dict from names to loci'''
        return self._dynamics.lociForProcess(self)

    def locus(self, n : str) -> Locus:
        '''Return the named locus.

        :param n: the locus name
        :returns: the locus'''
        return self.loci()[n]

    def addEventPerElement(self, l : Union[str, Locus], p : float, ef : EventFunction):
        """Add a probabilistic event at a locus, occurring with a particular (fixed)
        probability for each element of the locus, and calling the :term:`event function`
        when it is selected. The locus may be a :class:`Locus` object or a string, which
        is taken to be a locus of this process.

        Unlike fixed probability events added by :meth:`addFixedRateEvent`, a per-element
        event happens with the same probability to each element in the locus. This leads to
        larger loci giving rise to a higher rate of events.

        :param l: the locus or locus name
        :param p: the event probability
        :param ef: the event function"""
        self._dynamics.addEventPerElement(self, l, p, ef)

    def addFixedRateEvent(self, l : Union[str, Locus], p : float, ef : EventFunction):
        """Add a probabilistic event at a locus, occurring with a particular (fixed)
        probability, and calling the :term:`event function`
        when it is selected. The locus may be a :class:`Locus` object or a string, which
        is taken to be the name of a locus of this process.

        Unlike fixed rate events added by :meth:`addEventPerElement`, a fixed probability
        event happens with the same probability regardless of how many elements are in
        the locus.

        :param l: the locus or locus name
        :param p: the event probability
        :param ef: the event function"""
        self._dynamics.addFixedRateEvent(self, l, p, ef)

    def postEvent(self, t : float, e : Any, ef : EventFunction):
        """Post an event that calls the :term:`event function` at time t.

        :param t: the current time
        :param e: the element (node or edge) on which the event occurs
        :param ef: the event function"""
        self._dynamics.postEvent(t, e, ef)

    def postRepeatingEvent(self, t : float, dt : float, e : Any, ef : EventFunction):
        """Post an event that starts at time t and re-occurs at interval dt.

        :param t: the start time
        :param dt: the interval
        :param e: the element (node or edge) on which the event occurs
        :param ef: the element function"""
        self._dynamics.postRepeatingEvent(t, dt, e, ef)


    # ---------- Event distributions ----------

    def perElementEventDistribution(self, t : float) -> EventDistribution:
        """Return the distribution of per-element events at time t relevant to
        this process. The list only contains those loci that this process defined.

        The default is to use the distribution directly from the dynamics, which is
        time-independent. Sub-classes can extend this method to provide time dependency
        if needed.

        :param t: the simulation time
        :returns: a list of (locus, probability, event function) triples"""
        return self._dynamics.perElementEventDistribution(self)

    def fixedRateEventDistribution(self, t : float) -> EventDistribution:
        """Return the distribution of fixed-rate events at time t. relevant to
        this process. The list only contains those loci that this process defined.

        The default is to use the distribution directly from the dynamics, which is
        time-independent. Sub-classes can extend this method to provide time dependency
        if needed.

        :param t: the simulation time
        :returns: a list of (locus, probability, event function) triples"""
        return self._dynamics.fixedRateEventDistribution(self)

    def eventRateDistribution(self, t : float) -> EventDistribution:
        """Return the event distribution, a sequence of (l, r, f) triples
        where l is the locus where the event occurs, r is the rate at
        which an event occurs, and f is the event function called to
        make it happen.

        Note the distinction between a *rate* and a *probability*:
        the former can be obtained from the latter simply by
        multiplying the event probability by the number of times it's
        possible in the current network, which for per-element events
        is the population of nodes or edges in a given state.

        It is perfectly fine for an event to have a zero rate. The process
        is assumed to have reached equilibrium if all events have zero rates.

        :param t: current time
        :returns: a list of (locus, rate, event function) triples"""
        rates = []

        # convert per-element events to rates
        for (l, pr, ef) in self.perElementEventDistribution(t):
            rates.append((l, pr * len(l), ef))

        # add fixed-rate events for non-empty loci
        for (l, pr, ef) in self.fixedRateEventDistribution(t):
            if len(l) > 0:
                rates.append((l, pr, ef))
        return rates









# Network generator base class and basic instances
#
# Copyright (C) 2017--2020 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from networkx import Graph
from typing import Dict, Any, Optional, cast

class NetworkGenerator(object):
    '''An object that generates instances of networks on demand. Generators
    encapsulate the functions needed to create networks for experiments. These
    will typically be random, but might be loaded from a dataset.

    Sub-classes override the :meth:`_generate` method to create new instances
    of the network class described by this generator.

    A generator is driven by experimental parameters passed to the
    experiment represented by the :class:`Dynamics` class. These can be
    set at construction and/or re-set using the :meth:`set` method at
    any time.

    The maximum number of instances the generator can generate can also be limited,
    which essentially creates a set of networks of the given class that are
    generated on demand. The limit can't be changed and applies to the total number
    of networks generated, both singly (using :meth:`generate`) and by
    iteration (using :meth:`__iter__`).
    
    :param params: (optional) experimental parameters
    :param limit: (optional) maximum number of network instances to generate (defaults to unbounded)'''

    def __init__(self, params : Dict[str, Any] =None, limit : Optional[int] =None):
        if params is None:
            params = dict()
        self._params : Dict[str, Any] = cast(Dict[str, Any], params)
        self._remaining : Optional[int] = limit


    # ---------- Network generation ----------

    def set(self, params : Dict[str, Any]) -> 'NetworkGenerator':
        '''Set the parameters for the family of networks generated by this generator.

        :param params: the parameters
        :returns: the generator'''
        self._params = params.copy()
        return self

    def _generate(self, params : Dict[str, Any]) -> Graph:
        '''Generate an instance of a network. This method isn't called directly,
        but can be accessed through either the :meth:`generate` method or through
        the iterator interface, Sub-classes should override this method to provide 
        the code needed to create instances of the network family described by the
        generator, according to the experimental paramaters.

        :param params: the experimental parameters
        :returns: a network instance'''
        raise NotImplementedError('NetworkGenerator._generate needs to be overridden by sub-classes')

    def generate(self) -> Optional[Graph]:
        '''Generate a new instance of a network. The method returns
        None if the total number of allowed instances have already been generated.

        :returns: a network instance or None'''
        if self._remaining is None:
            return self._generate(self._params)
        elif self._remaining > 0:
            self._remaining -= 1
            return self._generate(self._params)
        else:
            return None


    # ---------- Iterator interface ----------

    def __iter__(self) -> 'NetworkGenerator':
        '''Return iterator over the family of networks this generator generates.

        :returns: the generator itself'''
        return self

    def __next__(self) -> Graph:
        '''Generate an instance of the network as long as the maximum number
        of instances (if any) have not already been generated.

        :returns: a network'''
        g = self.generate()
        if g is None:
            raise StopIteration()
        else:
            return g




# Test the behaviour of SIR over a dynamic population
#
# Copyright (C) 2017--2020 Simon Dobson
#
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *
import epyc
import unittest
import networkx

class DynamicSIR(SIR, AddDelete):

    N = 'networkSize'

    def __init__(self):
        super(DynamicSIR, self).__init__()

    def addNewNode(self, **kwds):
        '''Mark new nodes as susceptible.

        :param kwds: (optional) node attributes
        :returns: the generated name of the new node'''

        # add the node, capturing its name
        n = super(DynamicSIR, self).addNewNode(**kwds)

        # set the compartment of this node to susceptible
        self.setCompartment(n, SIR.SUSCEPTIBLE)

        # return the name of the new node
        return n

    def removeNode(self, n):
        '''Mark any node as removed before deleting.

        :param n: the node'''

        # change the node's compartment to removed
        self.changeCompartment(n, SIR.REMOVED)

        # delete the node
        super(DynamicSIR, self).removeNode(n)

    def results( self ):
        '''Save the size of the resulting network.

        :returns: a dict of experimental results'''
        rc = super(DynamicSIR, self).results()
        rc[self.N] = self.network().order()
        return rc


class AddDeleteSIRTest(unittest.TestCase):

    def setUp(self):
        '''Set up the experimental parameters and process.'''
        self._N = 5000
        self._kmean = 10
        self._phi = (self._kmean + 0.0) / self._N
        self._network = networkx.erdos_renyi_graph(self._N, self._phi)
        self._maxTime = 5000

        self._params = dict()
        self._params[AddDelete.DEGREE] = 10
        self._params[SIR.P_INFECT] = 0.1
        self._params[SIR.P_INFECTED] = 0.01
        self._params[SIR.P_REMOVE] = 0.05

        self._process = DynamicSIR()
        self._process.setMaximumTime(self._maxTime)
        self._e = StochasticDynamics(self._process, self._network)

    def testIndependent(self):
        '''Test that the epidemic doesn't affect the population size when run with equal rates.'''
        self._params[AddDelete.P_ADD] = 1
        self._params[AddDelete.P_DELETE] = 1
        rc = self._e.set(self._params).run()
        self.assertAlmostEqual(rc[epyc.Experiment.RESULTS][DynamicSIR.N], self._network.order(), delta = int((self._network.order() + 0.0) * 0.1))

    def testZeroRates(self):
        '''Test that zero rates leave a fixed population and a normal epidemic.'''
        self._params[AddDelete.P_ADD] = 0
        self._params[AddDelete.P_DELETE] = 0
        rc = self._e.set(self._params).run()
        self.assertEqual(rc[epyc.Experiment.RESULTS][DynamicSIR.N], self._network.order())
        self.assertEqual(rc[epyc.Experiment.RESULTS][DynamicSIR.N], self._N)
        self.assertCountEqual(rc[epyc.Experiment.RESULTS], [DynamicSIR.N, SIR.SUSCEPTIBLE, SIR.INFECTED, SIR.REMOVED])
        self.assertTrue(rc[epyc.Experiment.RESULTS][SIR.SUSCEPTIBLE] > 0)
        self.assertTrue(rc[epyc.Experiment.RESULTS][SIR.INFECTED] == 0)
        self.assertTrue(rc[epyc.Experiment.RESULTS][SIR.REMOVED] > 0)
        self.assertEqual(rc[epyc.Experiment.RESULTS][SIR.SUSCEPTIBLE] + rc[epyc.Experiment.RESULTS][SIR.REMOVED], self._network.order())

    def testAllInCompartments(self):
        '''Test that all nodes land in a compartment.'''
        self._params[AddDelete.P_ADD] = 1
        self._params[AddDelete.P_DELETE] = 1
        rc = self._e.set(self._params).run()
        self.assertEqual(rc[epyc.Experiment.RESULTS][DynamicSIR.N], rc[epyc.Experiment.RESULTS][SIR.SUSCEPTIBLE] + rc[epyc.Experiment.RESULTS][SIR.INFECTED] + rc[epyc.Experiment.RESULTS][SIR.REMOVED])

if __name__ == '__main__':
    unittest.main()